#ifndef LSM_STORE
#define LSM_STORE

#include <stdbool.h>
#include <stdint.h>

#include "lsm.h"
#include "lsm/str.h"

#define LSM_STORE_DISK_THRESHOLD 1024

/**
 * The type of an entry attribute.
 *
 * Each type is represented as a single bit of an
 * integer, so they can be easily combined into a bitmap.
 */
typedef enum lsm_attr_type : uint64_t {
  lsm_attr_type_entry_type = 1 << 0,
  lsm_attr_type_content_type = 1 << 1
} lsm_attr_type;

/**
 * An entry inside an LSM store.
 *
 * Each entry consists of the key it's stored behind, zero  or more attributes
 * (metadata) and a data field. The data field can be stored on disk or
 * in-memory, depending on the size.
 */
typedef struct lsm_entry lsm_entry;

/**
 * Allocate and initialize a new lsm_entry object.
 *
 * @param ptr where to store newly allocated pointer
 */
lsm_error lsm_entry_init(lsm_entry **ptr);

/**
 * Deallocate an existing lsm_entry object.
 *
 * @param entry object to deallocate
 */
void lsm_entry_free(lsm_entry *entry);

/**
 * Checks whether the entry has an attribute with the specified type.
 *
 * @param entry entry to check
 * @param type type of attribute to check for
 */
bool lsm_entry_attr_present(lsm_entry *entry, lsm_attr_type type);

/**
 * Retrieve the contents of an attribute from an entry, if present
 *
 * @param out where to store pointer to attribute data
 * @param entry entry to search for
 * @param type type of attribute to return
 */
lsm_error lsm_entry_attr_get(lsm_str **out, lsm_entry *entry,
                             lsm_attr_type type);

/**
 * Add a new attribute to the entry.
 *
 * @param entry entry to modify
 * @param type type of attribute to add
 * @param data data of attribute; ownership of pointer is taken over
 */
lsm_error lsm_entry_attr_insert(lsm_entry *entry, lsm_attr_type type,
                                lsm_str *data);

/**
 * Remove an atribute from the given entry, if present.
 *
 * @param out pointer to store removed data pointer in. If NULL, data pointer
 * will be leaked.
 * @param entry entry to remove attribute from
 * @param type type of attribute to remove
 */
lsm_error lsm_entry_attr_remove(lsm_str **out, lsm_entry *entry,
                                lsm_attr_type type);

/**
 * A store consisting of LSM entries.
 *
 * A store manages both an in-memory data structure for quick lookup, and a
 * database file for persistent storage of the contained entries.
 */
typedef struct lsm_store lsm_store;

/**
 * Allocate and initialize a new lsm_store object.
 *
 * @param ptr where to store newly allocated pointer
 */
lsm_error lsm_store_init(lsm_store **ptr);

/**
 * Open the given database file and load it into a new store object.
 *
 * @param ptr pointer to store newly allocated store
 * @param db_path path to the database file
 * @param data_path path to the data directory
 */
lsm_error lsm_store_open(lsm_store **ptr, lsm_str *db_path, lsm_str *data_path);

/**
 * Dealocate an existing lsm_store object.
 *
 * @param store object to deallocate
 */
void lsm_store_free(lsm_store *store);

/**
 * Retrieve an entry from the store, preparing & locking it for the purpose of
 * reading.
 *
 * @param out pointer to store entry pointer
 * @param store store to retrieve entry from
 * @param key key to search
 */
lsm_error lsm_store_get_read(lsm_entry **out, lsm_store *store, lsm_str *key);

/**
 * Retrieve an entry from the store for the purposes of writing. This
 * write-locks the entry.
 *
 * @param out pointer to store entry pointer
 * @param store store to retrieve entry from
 * @param key key to search
 */
lsm_error lsm_store_get_write(lsm_entry **out, lsm_store *store, lsm_str *key);

/**
 * Unlock a locked entry.
 *
 * @param store store to unlock entry in
 * @param entry entry to unlock
 */
lsm_error lsm_store_unlock(lsm_store *store, lsm_entry *entry);

/**
 * Allocate a new entry in the store with the specified key. The entry returned
 * will be write-locked, and should be unlocked after streaming the necessary
 * data.
 *
 * @param out pointer to store new entry pointer in
 * @param store store to modify
 * @param key key to add; ownership of key pointer is taken over
 */
lsm_error lsm_store_insert(lsm_entry **out, lsm_store *store, lsm_str *key);

/**
 * Append new data to the given entry, which is expected to be in the store.
 *
 * This function will append either to disk or to memory, depending on the
 * length of the entry's data.
 *
 * @param store store the entry is stored in
 * @param entry entry to append data to
 * @param data data to append
 */
lsm_error lsm_store_data_write(lsm_store *store, lsm_entry *entry,
                               lsm_str *data);

#endif
